@@ -6,29 +6,56 @@ import num2words
 CURRENCY_WORD_MAPPING = {
     "en": {"$": "dollar", "€": "euro", "£": "pound"},
     "de": {"$": "Dollar", "€": "Euro", "£": "Pfund"},
+    "fr": {"$": "dollar", "€": "euro", "£": "livre"},
+    "es": {"$": "dólar", "€": "euro", "£": "libra"},
 }
 
-INVALID_CHARS = {"en": r"[^a-zA-Z]", "de": r"[^a-zA-ZäöüÄÖÜß]"}
+INVALID_CHARS = {
+    "en": r"[^a-zA-Z]",
+    "de": r"[^a-zA-ZäöüÄÖÜß]",
+    "fr": r"[^a-zA-ZÀ-ÿ]",
+    "es": r"[^a-zA-ZÀ-ÿáéíñúüÁÉÍÑÓÚÜ]",
+}
 
-PERCENT_WORD = {"en": "percent", "de": "Prozent"}
-AND_WORD = {"en": "and", "de": "und"}
+PERCENT_WORD = {"en": "percent", "de": "Prozent", "fr": "pour cent", "es": "por ciento"}
+AND_WORD = {"en": "and", "de": "und", "fr": "et", "es": "y"}
 
 METRIC_WORD = {
     "en": {
         "km/h": "kilometers per hour",
         "mph": "miles per hour",
         "m/s": "meters per second",
+        "kg": "kilograms",
+        "cm": "centimeters",
     },
     "de": {
         "km/h": "Kilometer pro Stunde",
         "mph": "Meilen pro Stunde",
         "m/s": "Meter pro Sekunde",
+        "kg": "Kilogramm",
+        "cm": "Zentimeter",
+    },
+    "fr": {
+        "km/h": "kilomètres par heure",
+        "mph": "miles par heure",
+        "m/s": "mètres par seconde",
+        "kg": "kilogrammes",
+        "cm": "centimètres",
+    },
+    "es": {
+        "km/h": "kilómetros por hora",
+        "mph": "millas por hora",
+        "m/s": "metros por segundo",
+        "kg": "kilogramos",
+        "cm": "centímetros",
     },
 }
 
 YEAR_PRECEEDING_WORDS = {
     "en": ["in", "year", "before", "after", "since"],
     "de": ["im", "jahr", "vor", "ab", "seit"],
+    "fr": ["en", "année", "l'année", "avant", "depuis"],
+    "es": ["en", "año", "desde"],
 }
 
 # the dict key is regex
@@ -50,6 +77,7 @@ ABBR_WORDS = {
         r"\bmisc\.": "miscellaneous",
         r"\be\.g\.": "for example",
         r"\btv\b": "T V",
+        r"\bno\.": "number",
     },
     "de": {
         r"\bz\.B\. ": "zum Beispiel",
@@ -67,31 +95,60 @@ ABBR_WORDS = {
         r"\bMWSt\.": "Mehrwertsteuer",
         r"\bPLZ\b": "Postleitzahl",
         r"\bNK\b": "Nebenkosten",
+        r"\bNr\.": "Nummer",
+    },
+    "fr": {
+        r"\bAuj\. ": "aujourd’hui",
+        r"\bEx.\.": "exemple",
+        r"\bQQ\b": "quelque",
+        r"\bQQ’un\b": "quelqu'un",
+        r"\bS\.": "siècle",
+    },
+    "es": {
+        r"\bSra\. ": "señora",
+        r"\bSr\.": "señor",
+        r"\bSrta\.": "señorita",
+        r"\bDña\.": "doña",
+        r"\bS\.": "siècle",
     },
 }
 
-ABBR_SEARCH_FLAG = {"en": re.IGNORECASE, "de": 0}
+ABBR_SEARCH_FLAG = {
+    "en": re.IGNORECASE,
+    "de": 0,
+    "fr": re.IGNORECASE,
+    "es": re.IGNORECASE,
+}
 
 
 class BaseFormatter:
     def __init__(self, lang="en"):
-        if lang not in ["en", "de"]:
+        if lang not in ["en", "de", "fr", "es"]:
             raise Exception("language not implemented")
 
         self.lang = lang
 
 
 class TextFormatter(BaseFormatter):
-    def __init__(self, lang="en", return_upper_case=True):
+    def __init__(self, lang="en", return_case=0):
         super().__init__(lang)
         # the purpose is to remove ambiguity of cases
         # in German it might be unnecessary
-        self.return_upper_case = return_upper_case
+
+        """
+        return case:
+            0: no change
+            1: upper
+            -1: lower
+        """
+
+        self.return_case = return_case
 
         self.__cache = {}
 
     def n2w(self, word, to="cardinal"):
         word = re.sub(r"[^0-9]", "", word)
+        word = word[:15]  # 999 trillion most, should not handle too long digits
         numword = num2words.num2words(int(word), to=to, lang=self.lang)
         numword = numword.replace(",", "").replace("-", " ")
         return numword.split()
@@ -101,9 +158,15 @@ class TextFormatter(BaseFormatter):
         name = CURRENCY_WORD_MAPPING.get(self.lang).get(symbol)
         if name and re.match(r"^[\d,]+$", word[1:]):
             value = re.sub(r"[^0-9]", "", word[1:])  # return string
-            # TODO other languages
             if int(value) > 1:
-                name += "s"
+                if self.lang in ["en", "fr"]:
+                    name += "s"
+                elif self.lang == "es":
+                    if symbol == "$":
+                        name += "es"
+                    else:
+                        name += "s"
+                # German: no change
             return name, value
         return None
 
@@ -124,11 +187,7 @@ class TextFormatter(BaseFormatter):
         if c in string.punctuation:
             return False
 
-        if self.lang == "en":
-            return ord(c) < 128
-
-        if self.lang == "de":
-            return ord(c) < 128 or c in "äöüÄÖÜß"
+        return ord(c) < 128 or not re.match(INVALID_CHARS.get(self.lang), c)
 
     def contains_metric(self, word):
         w = word.lower()
@@ -144,6 +203,12 @@ class TextFormatter(BaseFormatter):
 
         return ending.upper()
 
+    def wrap_case(self, word):
+        if self.return_case == 0:
+            return word
+
+        return word.lower() if self.return_case == -1 else word.upper()
+
     def plural_ending(self, word):
         if self.lang == "en":
             last_lower = word[-1].lower()
@@ -227,7 +292,7 @@ class TextFormatter(BaseFormatter):
         search_flag = ABBR_SEARCH_FLAG[self.lang]
         for abbr, repl in ABBR_WORDS[self.lang].items():
             for found in re.findall(abbr, text, flags=search_flag):
-                change_list[found] = repl.upper() if self.return_upper_case else repl
+                change_list[found] = self.wrap_case(repl)
             text = re.sub(abbr, repl + " ", text, flags=search_flag)
 
         text = text.replace(".", " ")
@@ -342,17 +407,12 @@ class TextFormatter(BaseFormatter):
                         final_append.append(self.remove_invalid_chars(ta))
 
                 change_list[word] = " ".join(
-                    [
-                        x.upper() if self.return_upper_case else x
-                        for x in filter(len, final_append)
-                    ]
+                    [self.wrap_case(x) for x in filter(len, final_append)]
                 )
 
             word_list += final_append
 
-        text = " ".join(word_list)
-        if self.return_upper_case:
-            text = text.upper()
+        text = self.wrap_case(" ".join(word_list))
 
         self.__cache[original_text] = (text, change_list)
         return self.__cache[original_text]
@@ -380,6 +440,8 @@ class PhoneFormatter(BaseFormatter):
             )
             return phone_root
 
+        return phone.split(sep)[0]
+
     def get_phone_stress(self, phone, sep="_", return_detail=False):
         """
         0    — No stress
@@ -405,24 +467,45 @@ class PhoneFormatter(BaseFormatter):
         if not phone:
             return False
 
-        if self.lang == "en":
-            return phone[0] in ["A", "E", "I", "O", "U"]
+        phone_root = self.get_phone_root(phone)
 
-        if self.lang == "de":
-            return phone[0] in [
-                "a",
-                "e",
-                "E",
+        if self.lang == "en":
+            return phone_root[0] in ["A", "E", "I", "O", "U"]
+
+        # sampa common
+        # https://en.wikipedia.org/wiki/SAMPA_chart
+        if self.lang in ["de", "fr"]:
+            return phone_root in [
                 "i",
                 "I",
-                "o",
-                "O",
-                "u",
-                "U",
+                "e",
+                "E",
+                "{",
                 "y",
+                "Y",
                 "2",
                 "9",
+                "1",
                 "@",
+                "6",
+                "3",
+                "a",
+                "o",
+                "}",
+                "8",
+                "&",
+                "M",
+                "7",
+                "V",
+                "A",
+                "u",
+                "U",
+                "o",
+                "O",
+                "Q",
             ]
 
+        if self.lang == "es":
+            return phone_root[0] in ["a", "e", "i", "o", "u"]
+
         return False
